#include <BasicAbstractions/Assets.h>
#include <States/GameState/GameState.h>
#include <States/SelectState/SelectState.h>
#include <Themes/Themes.h>

GameState::GameState(IStateManager& state_manager, const sf::VideoMode& video_mode, const sf::String& window_title) :
        IState(state_manager), IWindowKeeper(video_mode, window_title) {
    m_window.setFramerateLimit(config::FRAME_RATE_LIMIT);
    m_window.setKeyRepeatEnabled(false);
    m_rect.setSize(config::NOTIFICATION_BANNER_SIZE);
    m_rect.setFillColor(sf::Color(240, 0, 0, 50));
    m_rect.setPosition(
        round((static_cast<float>(m_window.getSize().x) - m_rect.getSize().x) / 2.0f),
        round((static_cast<float>(m_window.getSize().y) - m_rect.getSize().y) / 2.0f)
    );
    m_text.setFont(Assets::GetFont());
    m_text.setCharacterSize(config::NOTIFICATION_FONT_SIZE);
    m_text.setStyle(sf::Text::Bold);
    m_text.setFillColor(Themes::Instance().get_text_color_notification());
}

void GameState::event_handling() {
    sf::Event event{};
    while (m_window.pollEvent(event)) {
        if (event.type == sf::Event::Closed) {
            m_state_manager.set_next_state(std::make_unique<SelectState>(m_state_manager, config::SELECT_LEVEL_VIDEO_MODE, config::SELECT_LEVEL_TITLE));
        }
        if (event.type == sf::Event::KeyPressed && event.key.code == sf::Keyboard::Space) {
            GameContext& context = m_context_manager.get_current_context();
            switch (context.state) {
                case GameContext::BEGIN:
                case GameContext::PAUSE:
                    context.state = GameContext::INGAME;
                    break;
                case GameContext::INGAME:
                    context.state = GameContext::PAUSE;
                    break;
                default: break;
            }
        }
        if (event.type == sf::Event::KeyPressed && event.key.code == sf::Keyboard::Z && event.key.control)
            m_context_manager.restore_previous_context();
        if (event.type == sf::Event::KeyPressed && m_context_manager.get_current_context().state == GameContext::INGAME)
            process_control_key_pressed(event.key.code);
        if (event.type == sf::Event::Resized) {
            sf::View view = m_window.getView();
            view.setSize(static_cast<float>(event.size.width), static_cast<float>(event.size.height));
            m_window.setView(view);
        }
    }
}

void GameState::update() {
    GameContext& current_context = m_context_manager.get_current_context();
    if (current_context.state == GameContext::LOST)
        m_progress_bar.update_defeat_progress_bar(current_context);
    if (current_context.state != GameContext::INGAME) return;
    std::vector<std::unique_ptr<IGameEvent>> events;
    for (const auto& obj : current_context.dynamic_objects) {
        obj->action();
        if (auto event_ptr = obj->accept(&current_context.pacman); event_ptr != nullptr)
            events.emplace_back(std::move(event_ptr));
    }
    for (const auto& obj : current_context.static_objects) {
        if (auto event_ptr = obj->accept(&current_context.pacman); event_ptr != nullptr)
            events.emplace_back(std::move(event_ptr));
    }
    if (current_context.static_objects.empty()) {
        events.emplace_back(std::make_unique<WinGame>());
    }
    current_context.pacman.move();
    for (const auto& event : events)
        event->handle(current_context);
    m_progress_bar.update_food_progress_bar(current_context);
}

void GameState::render() {
    draw_notification();
    m_maze.draw_into(m_window);
    for (const auto& obj : m_context_manager.get_current_context().static_objects)
        obj->draw_into(m_window);
    m_context_manager.get_current_context().pacman.draw_into(m_window);
    for (const auto& obj : m_context_manager.get_current_context().dynamic_objects)
        obj->draw_into(m_window);
    m_progress_bar.draw_into(m_window);
    if (m_context_manager.get_current_context().state != GameContext::INGAME) {
        m_window.draw(m_rect);
        m_window.draw(m_text);
    }
    m_window.display();
}

void GameState::draw_notification() {
        switch (m_context_manager.get_current_context().state) {
        case GameContext::BEGIN:
            m_window.clear(Themes::Instance().get_game_color_background_in_game());
            m_rect.setFillColor(Themes::Instance().get_notification_banner_color_play());
            m_text.setString(config::TEXT_PLAY);
            m_text.setPosition(
                m_rect.getPosition().x + (m_rect.getSize().x - m_text.getLocalBounds().width) / 2.0f - m_text.getLocalBounds().left,
                m_rect.getPosition().y + (m_rect.getSize().y - m_text.getLocalBounds().height) / 2.0f - m_text.getLocalBounds().top
            );
            break;
        case GameContext::INGAME:
            m_window.clear(Themes::Instance().get_game_color_background_in_game());
            break;
        case GameContext::PAUSE:
            m_window.clear(Themes::Instance().get_game_color_background_pause());
            m_rect.setFillColor(Themes::Instance().get_notification_banner_color_pause());
            m_text.setString(config::TEXT_PAUSE);
            m_text.setPosition(
                    m_rect.getPosition().x + (m_rect.getSize().x - m_text.getLocalBounds().width) / 2.0f - m_text.getLocalBounds().left,
                    m_rect.getPosition().y + (m_rect.getSize().y - m_text.getLocalBounds().height) / 2.0f - m_text.getLocalBounds().top
                );
            break;
        case GameContext::LOST:
            m_window.clear(Themes::Instance().get_game_color_background_lost());
            m_rect.setFillColor(Themes::Instance().get_notification_banner_color_lost());
            m_text.setString(config::TEXT_LOST);
            m_text.setPosition(
                    m_rect.getPosition().x + (m_rect.getSize().x - m_text.getLocalBounds().width) / 2.0f - m_text.getLocalBounds().left,
                    m_rect.getPosition().y + (m_rect.getSize().y - m_text.getLocalBounds().height) / 2.0f - m_text.getLocalBounds().top
                );
            break;
        case GameContext::WIN:
            m_window.clear(Themes::Instance().get_game_color_background_win());
            m_rect.setFillColor(Themes::Instance().get_notification_banner_color_win());
            m_text.setString(config::TEXT_WIN);
            m_text.setPosition(
                    m_rect.getPosition().x + (m_rect.getSize().x - m_text.getLocalBounds().width) / 2.0f - m_text.getLocalBounds().left,
                    m_rect.getPosition().y + (m_rect.getSize().y - m_text.getLocalBounds().height) / 2.0f - m_text.getLocalBounds().top
                );
            break;
    }
}

void GameState::process_control_key_pressed(const sf::Keyboard::Key key) {
    switch (key) {
        case config::KEY_UP:
        case config::KEY_UP_2:
            m_context_manager.save_context();
            m_context_manager.get_current_context().pacman.replace(Room::UP);
            break;
        case config::KEY_DOWN:
        case config::KEY_DOWN_2:
            m_context_manager.save_context();
            m_context_manager.get_current_context().pacman.replace(Room::DOWN);
            break;
        case config::KEY_LEFT:
        case config::KEY_LEFT_2:
            m_context_manager.save_context();
            m_context_manager.get_current_context().pacman.replace(Room::LEFT);
            break;
        case config::KEY_RIGHT:
        case config::KEY_RIGHT_2:
            m_context_manager.save_context();
            m_context_manager.get_current_context().pacman.replace(Room::RIGHT);
            break;
        default: break;
    }
}

bool GameState::do_step() {
    event_handling();
    update();
    render();
    return true;
}